/*
   This file provides functions to load/save
   zBoy parameters from/to the configuration
   file.

   This file is part of the zBoy project.

   Copyright (C) Mateusz Viste 2010-2012
*/

#include <stdio.h>
#include <stdint.h>
#include <string.h> /* strcat() */
#include "binary.h" /* provides convenience bxXXXXXXXX notations */
#include "getsavedir.h"
#include "zboystructs.h"



/* This routine checks the consistency of a zBoy configuration, and fixes it if needed */
void zBoy_fixconfig(struct zboyparamstype *zboyconf) {
  if ((zboyconf->ShowFPS != 0) && (zboyconf->ShowFPS != 1)) zboyconf->ShowFPS = 0;
  if ((zboyconf->NoScreenSaver != 0) && (zboyconf->NoScreenSaver != 1)) zboyconf->NoScreenSaver = 0;
  if ((zboyconf->ShowRomInfos != 0) && (zboyconf->ShowRomInfos != 1)) zboyconf->ShowRomInfos = 0;
  if ((zboyconf->NoSpeedLimit != 0) && (zboyconf->NoSpeedLimit != 1)) zboyconf->NoSpeedLimit = 0;
  if ((zboyconf->HiScoresMem != 0) && (zboyconf->HiScoresMem != 1)) zboyconf->HiScoresMem = 0;
  if ((zboyconf->NoCpuIdle != 0) && (zboyconf->NoCpuIdle != 1)) zboyconf->NoCpuIdle = 0;
  if ((zboyconf->fpslimit < 1) || (zboyconf->fpslimit > 60)) zboyconf->fpslimit = 60;
  if ((zboyconf->bootlogo != 0) && (zboyconf->bootlogo != 1)) zboyconf->bootlogo = 0;
  if ((zboyconf->palette != 0) && (zboyconf->palette != 1)) zboyconf->palette = 0;
  if ((zboyconf->colorization != 0) && (zboyconf->colorization != 1)) zboyconf->colorization = 1;
  switch (zboyconf->scalingalgo) {
    case REFRESHSCREEN_NOSCALE:
      if (zboyconf->GraphicScaleFactor != 1) zboyconf->GraphicScaleFactor = 1;
      break;
    case REFRESHSCREEN_BASICSCALE:
      if ((zboyconf->GraphicScaleFactor < 1) || (zboyconf->GraphicScaleFactor > 8)) zboyconf->GraphicScaleFactor = 4;
      break;
    case REFRESHSCREEN_2X:
    case REFRESHSCREEN_EAGLE:
      if ((zboyconf->GraphicScaleFactor != 2) && (zboyconf->GraphicScaleFactor != 4) && (zboyconf->GraphicScaleFactor != 6) && (zboyconf->GraphicScaleFactor != 8)) zboyconf->GraphicScaleFactor = 4;
      break;
    case REFRESHSCREEN_3X:
      if ((zboyconf->GraphicScaleFactor != 3) && (zboyconf->GraphicScaleFactor != 6)) zboyconf->GraphicScaleFactor = 3;
      break;
    default:
      zboyconf->scalingalgo = REFRESHSCREEN_2X;
      zboyconf->GraphicScaleFactor = 4;
      break;
  }
}


int zBoy_loadconfig(struct zboyparamstype *zboyconf) {
  char configfile[1024];
  uint8_t conftable[32], cksum = 5;
  FILE *conf;
  GetSaveDir(configfile);
  strcat(configfile, "zboy.dat");
  conf = fopen(configfile, "rb");
  if (conf != NULL) {
      int x;
      for (x = 0; x < 32; x++) {
        conftable[x] = fgetc(conf);
        if (x > 0) cksum += conftable[x];
      }
      fclose(conf);
      if (cksum == conftable[0]) { /* checksum verification */
          zboyconf->ShowFPS = (conftable[1] >> 7) & 1;
          zboyconf->NoScreenSaver = (conftable[1] >> 6) & 1;
          zboyconf->ShowRomInfos = (conftable[1] >> 5) & 1;
          zboyconf->NoSpeedLimit = (conftable[1] >> 4) & 1;
          zboyconf->HiScoresMem = (conftable[1] >> 3) & 1;
          zboyconf->NoCpuIdle = (conftable[1] >> 2) & 1;
          zboyconf->bootlogo = (conftable[1] >> 1) & 1;
          zboyconf->colorization = conftable[1] & 1;
          zboyconf->GraphicScaleFactor = conftable[2] & 15;
          zboyconf->scalingalgo = (conftable[2] >> 4) & 15;
          zboyconf->fpslimit = conftable[3];
          zboyconf->palette = conftable[4];
          /* now load key mappings */
          zboyconf->key_up = conftable[5];
          zboyconf->key_down = conftable[6];
          zboyconf->key_left = conftable[7];
          zboyconf->key_right = conftable[8];
          zboyconf->key_start = conftable[9];
          zboyconf->key_select = conftable[10];
          zboyconf->key_a = conftable[11];
          zboyconf->key_b = conftable[12];
          zboyconf->key_turboa = conftable[13];
          zboyconf->key_turbob = conftable[14];
          zboyconf->key_bckg = conftable[15];
          zboyconf->key_sprt = conftable[16];
          zboyconf->key_wind = conftable[17];
          zboyconf->key_save = conftable[18];
          zboyconf->key_load = conftable[19];
          zboyconf->key_asht = conftable[20];
          zboyconf->key_shot = conftable[21];
          zboyconf->key_rset = conftable[22];
          zboyconf->key_quit = conftable[23];
	  /* all done */
          return(0);
        } else {
          return(-2);
      }
    } else {
      return(-1);
  }
}


int zBoy_saveconfig(struct zboyparamstype *zboyconf) {
  char configfile[1024];
  uint8_t conftable[32] = {5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  FILE *conf;
  GetSaveDir(configfile);
  strcat(configfile, "zboy.dat");
  conf = fopen(configfile, "wb");
  if (conf != NULL) {
      int x;
      if (zboyconf->ShowFPS != 0) conftable[1] |= bx10000000;
      if (zboyconf->NoScreenSaver != 0) conftable[1] |= bx01000000;
      if (zboyconf->ShowRomInfos != 0) conftable[1] |= bx00100000;
      if (zboyconf->NoSpeedLimit != 0) conftable[1] |= bx00010000;
      if (zboyconf->HiScoresMem != 0) conftable[1] |= bx00001000;
      if (zboyconf->NoCpuIdle != 0) conftable[1] |= bx00000100;
      if (zboyconf->bootlogo != 0) conftable[1] |= bx00000010;
      if (zboyconf->colorization != 0) conftable[1] |= bx00000001;
      conftable[2] = (zboyconf->scalingalgo << 4);
      conftable[2] |= zboyconf->GraphicScaleFactor;
      conftable[3] = zboyconf->fpslimit;
      conftable[4] = zboyconf->palette;
      /* now save key mappings */
      conftable[5] = zboyconf->key_up;
      conftable[6] = zboyconf->key_down;
      conftable[7] = zboyconf->key_left;
      conftable[8] = zboyconf->key_right;
      conftable[9] = zboyconf->key_start;
      conftable[10] = zboyconf->key_select;
      conftable[11] = zboyconf->key_a;
      conftable[12] = zboyconf->key_b;
      conftable[13] = zboyconf->key_turboa;
      conftable[14] = zboyconf->key_turbob;
      conftable[15] = zboyconf->key_bckg;
      conftable[16] = zboyconf->key_sprt;
      conftable[17] = zboyconf->key_wind;
      conftable[18] = zboyconf->key_save;
      conftable[19] = zboyconf->key_load;
      conftable[20] = zboyconf->key_asht;
      conftable[21] = zboyconf->key_shot;
      conftable[22] = zboyconf->key_rset;
      conftable[23] = zboyconf->key_quit;
      /* compute checkcum, write to file and quit */
      for (x = 1; x < 32; x++) conftable[0] += conftable[x]; /* compute the checksum */
      for (x = 0; x < 32; x++) fprintf(conf, "%c", conftable[x]); /* write to file */
      fclose(conf);
      return(0);
    } else {
      return(-1);
  }
}
