/*
    Joypad emulation of a GameBoy console
    Part of the zBoy project
    Copyright (C) Mateusz Viste 2010, 2011

   The joypad is controlled via the register at $FF00:
    Bit 5 - P15 out
    Bit 4 - P14 out
    Bit 3 - P13 in
    Bit 2 - P12 in
    Bit 1 - P11 in
    Bit 0 - P10 in

    When one of the keys gets a transition from high to low, an interrupt is fired.
    I got the list of SDL keycodes from http://sdl.beuc.net/sdl.wiki/SDLKey
*/

struct KeyDefinition { /* Used by keymap definitions */
  int32_t KeyCode;    /* SDL keyb scancode */
  int32_t JoyId;
  int32_t JoyBut;   /* button id */
  int32_t JoyAxisId;  /* axis number (0..3) */
  int32_t JoyAxisSig;  /* axis signum (-1 or +1) */
};

struct KeymapDefinition { /* Used for keymap definition */
  struct KeyDefinition Up;
  struct KeyDefinition Down;
  struct KeyDefinition Left;
  struct KeyDefinition Right;
  struct KeyDefinition Start;
  struct KeyDefinition Select;
  struct KeyDefinition A;
  struct KeyDefinition B;
  struct KeyDefinition TurboA;
  struct KeyDefinition TurboB;
  struct KeyDefinition bgCtrl;
  struct KeyDefinition spriteCtrl;
  struct KeyDefinition windowCtrl;
  struct KeyDefinition Load;
  struct KeyDefinition Save;
  struct KeyDefinition Autoshot;
  struct KeyDefinition Screenshot;
  struct KeyDefinition Reset;
  struct KeyDefinition Quit;
};

struct KeyStateMap {  /* Used for internal monitoring of key states */
  int32_t Up;
  int32_t Down;
  int32_t Right;
  int32_t Left;
  int32_t Select;
  int32_t Start;
  int32_t A;
  int32_t B;
};

uint32_t JoyOldReg = 0;

uint8_t ThereIsAJoypad = 0, ThereIsASecondJoypad = 0;
struct KeyStateMap KeyState;
struct KeymapDefinition KeyDef;


static char *keys[121] = {"none", "kb.f1", "kb.f2", "kb.f3", "kb.f4", "kb.f5", "kb.f6", "kb.f7", "kb.f8", "kb.f9",
"kb.f10", "kb.f11", "kb.f12", "kb.esc", "kb.tab", "kb.lctrl", "kb.lalt", "kb.rctrl", "kb.ralt", "kb.spc",
"kb.ret", "kb.bks", "kb.left", "kb.right", "kb.up", "kb.down", "kb.rshift", "kb.lshift", "kb.lwin", "kb.rwin",
"kb.menu", "kb.pgup", "kb.pgdw", "kb.home", "kb.end", "kb.ins", "kb.del", "kb.a", "kb.b", "kb.c",
"kb.d", "kb.e", "kb.f", "kb.g", "kb.h", "kb.i", "kb.j", "kb.k", "kb.l", "kb.m",
"kb.n", "kb.o", "kb.p", "kb.q", "kb.r", "kb.s", "kb.t", "kb.u", "kb.v", "kb.w",
"kb.x", "kb.y", "kb.z", "kb.0", "kb.1", "kb.2", "kb.3", "kb.4", "kb.5", "kb.6",
"kb.7", "kb.8", "kb.9", "joy1.a1pos", "joy1.a1neg", "joy1.a2pos", "joy1.a2neg", "joy1.a3pos", "joy1.a3neg", "joy1.a4pos",
"joy1.a4neg", "joy1.b1", "joy1.b2", "joy1.b3", "joy1.b4", "joy1.b5", "joy1.b6", "joy1.b7", "joy1.b8", "joy1.b9",
"joy1.b10", "joy1.b11", "joy1.b12", "joy1.b13", "joy1.b14", "joy1.b15", "joy1.b16", "joy2.a1pos", "joy2.a1neg", "joy2.a2pos",
"joy2.a2neg", "joy2.a3pos", "joy2.a3neg", "joy2.a4pos", "joy2.a4neg", "joy2.b1", "joy2.b2", "joy2.b3", "joy2.b4", "joy2.b5",
"joy2.b6", "joy2.b7", "joy2.b8", "joy2.b9", "joy2.b10", "joy2.b11", "joy2.b12", "joy2.b13", "joy2.b14", "joy2.b15",
"joy2.b16"};

int keystring2int(char *keystring) {
  int x;
  for (x = 0; x < 121; x++) if (strcasecmp(keys[x],keystring) == 0) return(x);
  return(0); /* else return 'none' */
}

/* returns a pointer to a string, or NULL if keycode is invalid */
char *int2keystring(int keycode) {
  if ((keycode >= 0) && (keycode < 121)) return(keys[keycode]); else return(NULL);
}



uint8_t zboyScancodeToBackendScancode(char *zboyScanCode, struct KeyDefinition *KeydefVar) {
  if (strcasecmp(zboyScanCode, "none") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f1") == 0) {
    KeydefVar->KeyCode = SDLK_F1;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f2") == 0) {
    KeydefVar->KeyCode = SDLK_F2;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f3") == 0) {
    KeydefVar->KeyCode = SDLK_F3;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f4") == 0) {
    KeydefVar->KeyCode = SDLK_F4;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f5") == 0) {
    KeydefVar->KeyCode = SDLK_F5;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f6") == 0) {
    KeydefVar->KeyCode = SDLK_F6;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f7") == 0) {
    KeydefVar->KeyCode = SDLK_F7;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f8") == 0) {
    KeydefVar->KeyCode = SDLK_F8;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f9") == 0) {
    KeydefVar->KeyCode = SDLK_F9;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f10") == 0) {
    KeydefVar->KeyCode = SDLK_F10;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f11") == 0) {
    KeydefVar->KeyCode = SDLK_F11;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f12") == 0) {
    KeydefVar->KeyCode = SDLK_F12;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.esc") == 0) {
    KeydefVar->KeyCode = SDLK_ESCAPE;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.tab") == 0) {
    KeydefVar->KeyCode = SDLK_TAB;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.lctrl") == 0) {
    KeydefVar->KeyCode = SDLK_LCTRL;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.rctrl") == 0) {
    KeydefVar->KeyCode = SDLK_RCTRL;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.lalt") == 0) {
    KeydefVar->KeyCode = SDLK_LALT;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.ralt") == 0) {
    KeydefVar->KeyCode = SDLK_RALT;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.spc") == 0) {
    KeydefVar->KeyCode = SDLK_SPACE;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.ret") == 0) {
    KeydefVar->KeyCode = SDLK_RETURN;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.bks") == 0) {
    KeydefVar->KeyCode = SDLK_BACKSPACE;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.left") == 0) {
    KeydefVar->KeyCode = SDLK_LEFT;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.right") == 0) {
    KeydefVar->KeyCode = SDLK_RIGHT;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.up") == 0) {
    KeydefVar->KeyCode = SDLK_UP;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  } 
  if (strcasecmp(zboyScanCode, "kb.down") == 0) {
    KeydefVar->KeyCode = SDLK_DOWN;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.rshift") == 0) {
    KeydefVar->KeyCode = SDLK_RSHIFT;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.lshift") == 0) {
    KeydefVar->KeyCode = SDLK_LSHIFT;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.lwin") == 0) {
    KeydefVar->KeyCode = SDLK_LSUPER;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.rwin") == 0) {
    KeydefVar->KeyCode = SDLK_RSUPER;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.menu") == 0) {
    KeydefVar->KeyCode = SDLK_MENU;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.pgup") == 0) {
    KeydefVar->KeyCode = SDLK_PAGEUP;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.pgdw") == 0) {
    KeydefVar->KeyCode = SDLK_PAGEDOWN;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.home") == 0) {
    KeydefVar->KeyCode = SDLK_HOME;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.end") == 0) {
    KeydefVar->KeyCode = SDLK_END;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.ins") == 0) {
    KeydefVar->KeyCode = SDLK_INSERT;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.del") == 0) {
    KeydefVar->KeyCode = SDLK_DELETE;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.a") == 0) {
    KeydefVar->KeyCode = SDLK_a;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.b") == 0) {
    KeydefVar->KeyCode = SDLK_b;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.c") == 0) {
    KeydefVar->KeyCode = SDLK_c;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.d") == 0) {
    KeydefVar->KeyCode = SDLK_d;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.e") == 0) {
    KeydefVar->KeyCode = SDLK_e;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.f") == 0) {
    KeydefVar->KeyCode = SDLK_f;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.g") == 0) {
    KeydefVar->KeyCode = SDLK_g;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.h") == 0) {
    KeydefVar->KeyCode = SDLK_h;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.i") == 0) {
    KeydefVar->KeyCode = SDLK_i;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.j") == 0) {
    KeydefVar->KeyCode = SDLK_j;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.k") == 0) {
    KeydefVar->KeyCode = SDLK_k;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.l") == 0) {
    KeydefVar->KeyCode = SDLK_l;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.m") == 0) {
    KeydefVar->KeyCode = SDLK_m;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.n") == 0) {
    KeydefVar->KeyCode = SDLK_n;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.o") == 0) {
    KeydefVar->KeyCode = SDLK_o;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.p") == 0) {
    KeydefVar->KeyCode = SDLK_p;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.q") == 0) {
    KeydefVar->KeyCode = SDLK_q;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.r") == 0) {
    KeydefVar->KeyCode = SDLK_r;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.s") == 0) {
    KeydefVar->KeyCode = SDLK_s;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.t") == 0) {
    KeydefVar->KeyCode = SDLK_t;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.u") == 0) {
    KeydefVar->KeyCode = SDLK_u;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.v") == 0) {
    KeydefVar->KeyCode = SDLK_v;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.w") == 0) {
    KeydefVar->KeyCode = SDLK_w;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.x") == 0) {
    KeydefVar->KeyCode = SDLK_x;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.y") == 0) {
    KeydefVar->KeyCode = SDLK_y;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.z") == 0) {
    KeydefVar->KeyCode = SDLK_z;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.0") == 0) {
    KeydefVar->KeyCode = SDLK_0;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.1") == 0) {
    KeydefVar->KeyCode = SDLK_1;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.2") == 0) {
    KeydefVar->KeyCode = SDLK_2;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.3") == 0) {
    KeydefVar->KeyCode = SDLK_3;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.4") == 0) {
    KeydefVar->KeyCode = SDLK_4;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.5") == 0) {
    KeydefVar->KeyCode = SDLK_5;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.6") == 0) {
    KeydefVar->KeyCode = SDLK_6;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.7") == 0) {
    KeydefVar->KeyCode = SDLK_7;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.8") == 0) {
    KeydefVar->KeyCode = SDLK_8;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "kb.9") == 0) {
    KeydefVar->KeyCode = SDLK_9;    /* SDL keyb scancode */
    KeydefVar->JoyId = -1;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a1pos") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 0;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = 1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a1neg") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 0;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a2pos") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = 1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a2neg") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a3pos") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 2;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = 1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a3neg") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 2;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a4pos") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 3;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = 1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.a4neg") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = -1;   /* button id */
    KeydefVar->JoyAxisId = 3;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b1") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 0;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b2") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 1;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b3") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 2;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b4") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 3;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b5") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 4;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b6") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 5;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b7") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 6;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b8") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 7;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b9") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 8;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b10") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 9;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b11") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 10;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b12") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 11;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b13") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 12;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b14") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 13;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b15") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 14;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  if (strcasecmp(zboyScanCode, "joy1.b16") == 0) {
    KeydefVar->KeyCode = -1;    /* SDL keyb scancode */
    KeydefVar->JoyId = 0;
    KeydefVar->JoyBut = 15;   /* button id */
    KeydefVar->JoyAxisId = -1;  /* axis number (0..3) */
    KeydefVar->JoyAxisSig = -1;  /* axis signum (-1 or +1) */
    return(0);
  }
  /*
    CASE "joy2.a1pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 1
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a1neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 1
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.a2pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 2
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a2neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 2
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.a3pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 3
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a3neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 3
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.a4pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 4
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a4neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 4
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.b1"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000000001
      ThereIsASecondJoypad = 1
    CASE "joy2.b2"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000000010
      ThereIsASecondJoypad = 1
    CASE "joy2.b3"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000000100
      ThereIsASecondJoypad = 1
    CASE "joy2.b4"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000001000
      ThereIsASecondJoypad = 1
    CASE "joy2.b5"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000010000
      ThereIsASecondJoypad = 1
    CASE "joy2.b6"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000100000
      ThereIsASecondJoypad = 1
    CASE "joy2.b7"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000001000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b8"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000010000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b9"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000100000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b10"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000001000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b11"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000010000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b12"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000100000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b13"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0001000000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b14"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0010000000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b15"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0100000000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b16"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b1000000000000000
      ThereIsASecondJoypad = 1 */
  return(1);
}


int RedefineInputKeys(char *sKeyUp, char *sKeyDown, char *sKeyLeft, char *sKeyRight, char *sKeyStart, char *sKeySelect, char *sKeyA, char *sKeyB, char *sKeyTurboA, char *sKeyTurboB) {
  if (zboyScancodeToBackendScancode(sKeyUp, &KeyDef.Up) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyDown, &KeyDef.Down) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyLeft, &KeyDef.Left) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyRight, &KeyDef.Right) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyStart, &KeyDef.Start) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeySelect, &KeyDef.Select) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyA, &KeyDef.A) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyB, &KeyDef.B) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyTurboA, &KeyDef.TurboA) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyTurboB, &KeyDef.TurboB) != 0) return(1);
  return(0);
}


int RedefineFunctionKeys(char *sKeyBg, char *sKeySprite, char *sKeyWindow, char *sKeySave, char *sKeyLoad, char *sKeyAutoshot, char *sKeyShot, char *sKeyRst, char *sKeyQuit) {
  if (zboyScancodeToBackendScancode(sKeyBg, &KeyDef.bgCtrl) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeySprite, &KeyDef.spriteCtrl) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyWindow, &KeyDef.windowCtrl) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeySave, &KeyDef.Save) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyLoad, &KeyDef.Load) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyAutoshot, &KeyDef.Autoshot) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyShot, &KeyDef.Screenshot) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyRst, &KeyDef.Reset) != 0) return(1);
  if (zboyScancodeToBackendScancode(sKeyQuit, &KeyDef.Quit) != 0) return(1);
  return(0);
}


/* Request a joypad INT */
#define CallJoyINT() IoRegisters[0xFF0F] |= bx00010000

#define DownIsPressed(); \
  if (KeyState.Down == 0) { \
    KeyState.Down = 1; \
    CallJoyINT(); \
  }

#define DownIsReleased(); \
  KeyState.Down = 0;

#define UpIsPressed(); \
  if (KeyState.Up == 0) { \
    KeyState.Up = 1; \
    CallJoyINT(); \
  }

#define UpIsReleased(); \
  KeyState.Up = 0;

#define LeftIsPressed(); \
  if (KeyState.Left == 0) { \
    KeyState.Left = 1; \
    CallJoyINT(); \
  }

#define LeftIsReleased(); \
  KeyState.Left = 0;

#define RightIsPressed(); \
  if (KeyState.Right == 0) { \
    KeyState.Right = 1; \
    CallJoyINT(); \
  }

#define RightIsReleased(); \
  KeyState.Right = 0;

#define StartIsPressed(); \
  if (KeyState.Start == 0) { \
    KeyState.Start = 1; \
    CallJoyINT(); \
  }

#define StartIsReleased(); \
  KeyState.Start = 0;

#define SelectIsPressed(); \
  if (KeyState.Select == 0) { \
    KeyState.Select = 1; \
    CallJoyINT(); \
  }

#define SelectIsReleased(); \
  KeyState.Select = 0;

#define ButtonAisPressed(); \
  if (KeyState.A == 0) { \
    KeyState.A = 1; \
    CallJoyINT(); \
  }

#define ButtonAisReleased(); \
  KeyState.A = 0;

#define ButtonBisPressed(); \
  if (KeyState.B == 0) { \
    KeyState.B = 1; \
    CallJoyINT(); \
  }

#define ButtonBisReleased(); \
  KeyState.B = 0;

#define ButtonTurboAisPressed(); \
  turbo_a = 1; \
  turbo_a_nextswap = TotalCycles;

#define ButtonTurboAisReleased(); \
  turbo_a = 0; \
  KeyState.A = 0;

#define ButtonTurboBisPressed(); \
  turbo_b = 1; \
  turbo_b_nextswap = TotalCycles;

#define ButtonTurboBisReleased(); \
  turbo_b = 0; \
  KeyState.B = 0;

#define ButtonBgctrlIsReleased(); \
  if (HideBackgroundDisplay == 0) { \
      HideBackgroundDisplay = 1; \
      SetUserMsg("BACKGROUND OFF"); \
    } else { \
      HideBackgroundDisplay = 0; \
      SetUserMsg("BACKGROUND ON"); \
  }

#define ButtonSprctrlIsReleased(); \
  if (HideSpritesDisplay == 0) { \
      HideSpritesDisplay = 1; \
      SetUserMsg("SPRITES OFF"); \
    } else { \
      HideSpritesDisplay = 0; \
      SetUserMsg("SPRITES ON"); \
  }

#define ButtonWinctrlIsReleased(); \
  if (HideWindowDisplay == 0) { \
      HideWindowDisplay = 1; \
      SetUserMsg("WINDOW OFF"); \
    } else { \
      HideWindowDisplay = 0; \
      SetUserMsg("WINDOW ON"); \
  }

#define ButtonScreenshotIsReleased(); \
  AskForScreenshot = 1;

#define ButtonAutoscreenshotIsReleased(); \
  if (AutoScreenshot == 0) { \
      AutoScreenshot = 1; \
      SetUserMsg("AUTOSCREENSHOT ON"); \
    } else { \
      AutoScreenshot = 0; \
      SetUserMsg("AUTOSCREENSHOT OFF"); \
  }

#define ButtonResetIsReleased(); \
  SetUserMsg("RESET..."); \
  if (zboyparams->HiScoresMem != 0) SaveHiScoresMem(); \
  ResetEmulator(); \
  if (zboyparams->HiScoresMem != 0) LoadHiScoresMem(); \

#define ButtonSaveIsReleased(); \
  SaveGame();

#define ButtonLoadIsReleased(); \
  LoadGame();

#define ButtonQuitIsReleased(); \
  QuitEmulator = 1;


inline void CheckJoypad(int cycles, struct zboyparamstype *zboyparams) {
  static uint64_t JoyCheckCounter = 0;
  static unsigned int turbo_a = 0, turbo_b = 0;
  static unsigned int JoyNewReg;
  static uint64_t turbo_a_nextswap = 0, turbo_b_nextswap = 0;
  static SDL_Event event;

  JoyCheckCounter += cycles;

  if (JoyCheckCounter > 80000) {  /* Check joypad state every ~20 ms */
    JoyCheckCounter = 0;
    /* Now poll the SDL event queue... */
    while (SDL_PollEvent(&event)) {
        switch (event.type) {
            case SDL_KEYDOWN:
                if (event.key.keysym.sym == KeyDef.Start.KeyCode) {
                    StartIsPressed();
                  } else if (event.key.keysym.sym == KeyDef.Select.KeyCode) {
                    SelectIsPressed();
                  } else if (event.key.keysym.sym == KeyDef.B.KeyCode) {
                    ButtonBisPressed();
                  } else if (event.key.keysym.sym == KeyDef.A.KeyCode) {
                    ButtonAisPressed();
                  } else if (event.key.keysym.sym == KeyDef.TurboA.KeyCode) {
                    ButtonTurboAisPressed();
                  } else if (event.key.keysym.sym == KeyDef.TurboB.KeyCode) {
                    ButtonTurboBisPressed();
                  } else if (event.key.keysym.sym == KeyDef.Up.KeyCode) {
                    UpIsPressed();
                  } else if (event.key.keysym.sym == KeyDef.Down.KeyCode) {
                    DownIsPressed();
                  } else if (event.key.keysym.sym == KeyDef.Left.KeyCode) {
                    LeftIsPressed();
                  } else if (event.key.keysym.sym == KeyDef.Right.KeyCode) {
                    RightIsPressed();
                }
                break;
            case SDL_KEYUP:
                if (event.key.keysym.sym == KeyDef.Start.KeyCode) { /* ENTER */
                    StartIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Select.KeyCode) { /* TAB */
                    SelectIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.B.KeyCode) { /* ALT */
                    ButtonBisReleased();
                  } else if (event.key.keysym.sym == KeyDef.A.KeyCode) { /* LCTRL */
                    ButtonAisReleased();
                  } else if (event.key.keysym.sym == KeyDef.TurboA.KeyCode) {
                    ButtonTurboAisReleased();
                  } else if (event.key.keysym.sym == KeyDef.TurboB.KeyCode) {
                    ButtonTurboBisReleased();
                  } else if (event.key.keysym.sym == KeyDef.Up.KeyCode) { /* UP */
                    UpIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Down.KeyCode) { /* DOWN */
                    DownIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Left.KeyCode) { /* LEFT */
                    LeftIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Right.KeyCode) { /* RIGHT */
                    RightIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.bgCtrl.KeyCode) { /* F1 */
                    ButtonBgctrlIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.spriteCtrl.KeyCode) { /* F2 */
                    ButtonSprctrlIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.windowCtrl.KeyCode) { /* F3 */
                    ButtonWinctrlIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Save.KeyCode) { /* F5 */
                    ButtonSaveIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Load.KeyCode) { /* F7 */
                    ButtonLoadIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Autoshot.KeyCode) { /* F8 */
                    ButtonAutoscreenshotIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Screenshot.KeyCode) { /* F9 */
                    ButtonScreenshotIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Reset.KeyCode) { /* F10 */
                    ButtonResetIsReleased();
                  } else if (event.key.keysym.sym == KeyDef.Quit.KeyCode) { /* ESC */
                    ButtonQuitIsReleased();
                }
                break;
            case SDL_QUIT:
                QuitEmulator = 1;
                break;
            case SDL_JOYBUTTONDOWN:  /* Handle Joystick Button Presses */
              if (event.jbutton.button == KeyDef.A.JoyBut) {
                  ButtonAisPressed();
                } else if (event.jbutton.button == KeyDef.B.JoyBut) {
                  ButtonBisPressed();
                } else if (event.jbutton.button == KeyDef.Start.JoyBut) {
                  StartIsPressed();
                } else if (event.jbutton.button == KeyDef.Select.JoyBut) {
                  SelectIsPressed();
                } else if (event.jbutton.button == KeyDef.TurboA.JoyBut) {
                  ButtonTurboAisPressed();
                } else if (event.jbutton.button == KeyDef.TurboB.JoyBut) {
                  ButtonTurboBisPressed();
                } else if (event.jbutton.button == KeyDef.Up.JoyBut) {
                  UpIsPressed();
                } else if (event.jbutton.button == KeyDef.Down.JoyBut) {
                  DownIsPressed();
                } else if (event.jbutton.button == KeyDef.Left.JoyBut) {
                  LeftIsPressed();
                } else if (event.jbutton.button == KeyDef.Right.JoyBut) {
                  RightIsPressed();
              }
              break;
            case SDL_JOYBUTTONUP:  /* Handle Joystick Button Releases */
              if (event.jbutton.button == KeyDef.A.JoyBut) {
                  ButtonAisReleased();
                } else if (event.jbutton.button == KeyDef.B.JoyBut) {
                  ButtonBisReleased();
                } else if (event.jbutton.button == KeyDef.Start.JoyBut) {
                  StartIsReleased();
                } else if (event.jbutton.button == KeyDef.Select.JoyBut) {
                  SelectIsReleased();
                } else if (event.jbutton.button == KeyDef.TurboA.JoyBut) {
                  ButtonTurboAisReleased();
                } else if (event.jbutton.button == KeyDef.TurboB.JoyBut) {
                  ButtonTurboBisReleased();
                } else if (event.jbutton.button == KeyDef.Up.JoyBut) {
                  UpIsReleased();
                } else if (event.jbutton.button == KeyDef.Down.JoyBut) {
                  DownIsReleased();
                } else if (event.jbutton.button == KeyDef.Left.JoyBut) {
                  LeftIsReleased();
                } else if (event.jbutton.button == KeyDef.Right.JoyBut) {
                  RightIsReleased();
                } else if (event.jbutton.button == KeyDef.bgCtrl.JoyBut) { /* F1 */
                  ButtonBgctrlIsReleased();
                } else if (event.jbutton.button == KeyDef.spriteCtrl.JoyBut) { /* F2 */
                  ButtonSprctrlIsReleased();
                } else if (event.jbutton.button == KeyDef.windowCtrl.JoyBut) { /* F3 */
                  ButtonWinctrlIsReleased();
                } else if (event.jbutton.button == KeyDef.Save.JoyBut) { /* F5 */
                  ButtonSaveIsReleased();
                } else if (event.jbutton.button == KeyDef.Load.JoyBut) { /* F7 */
                  ButtonLoadIsReleased();
                } else if (event.jbutton.button == KeyDef.Autoshot.JoyBut) { /* F8 */
                  ButtonAutoscreenshotIsReleased();
                } else if (event.jbutton.button == KeyDef.Screenshot.JoyBut) { /* F9 */
                  ButtonScreenshotIsReleased();
                } else if (event.jbutton.button == KeyDef.Reset.JoyBut) { /* F10 */
                  ButtonResetIsReleased();
                } else if (event.jbutton.button == KeyDef.Quit.JoyBut) { /* ESC */
                  ButtonQuitIsReleased();
              }
              break;
            case SDL_JOYAXISMOTION:
              if (event.jaxis.axis == KeyDef.Right.JoyAxisId) {
                if (((event.jaxis.value > 3200) && (KeyDef.Right.JoyAxisSig > 0)) || ((event.jaxis.value < -3200) && (KeyDef.Right.JoyAxisSig < 0))) {  /* right */
                    if (KeyState.Right == 0) RightIsPressed();
                  } else {
                    KeyState.Right = 0;
                }
              }
              if (event.jaxis.axis == KeyDef.Left.JoyAxisId) {
                if (((event.jaxis.value > 3200) && (KeyDef.Left.JoyAxisSig > 0)) || ((event.jaxis.value < -3200) && (KeyDef.Left.JoyAxisSig < 0))) {  /* left */
                    if (KeyState.Left == 0) LeftIsPressed();
                  } else {
                    KeyState.Left = 0;
                }
              }
              if (event.jaxis.axis == KeyDef.Down.JoyAxisId) {
                if (((event.jaxis.value > 3200) && (KeyDef.Down.JoyAxisSig > 0)) || ((event.jaxis.value < -3200) && (KeyDef.Down.JoyAxisSig < 0))) {  /* down */
                    if (KeyState.Down == 0) DownIsPressed();
                  } else {
                    KeyState.Down = 0;
                }
              }
              if (event.jaxis.axis == KeyDef.Up.JoyAxisId) {
                if (((event.jaxis.value > 3200) && (KeyDef.Up.JoyAxisSig > 0)) || ((event.jaxis.value < -3200) && (KeyDef.Up.JoyAxisSig < 0))) {  /* up */
                    if (KeyState.Up == 0) UpIsPressed();
                  } else {
                    KeyState.Up = 0;
                }
              }
              break;
        }
    }

    /* Let's see if we have to deal with turbo keys now */
    if (turbo_a != 0) {
      if (turbo_a_nextswap < TotalCycles) {
        turbo_a_nextswap = TotalCycles + 400000; /* add 0.1s */
        if (KeyState.A != 0) {
            KeyState.A = 0;
          } else {
            ButtonAisPressed();
        }
      }
    }
    if (turbo_b != 0) {
      if (turbo_b_nextswap < TotalCycles) {
        turbo_b_nextswap = TotalCycles + 400000;
        if (KeyState.B != 0) {
            KeyState.B = 0;
          } else {
            ButtonBisPressed();
        }
      }
    }
  } /* if (JoyCheckCounter > 80000) */

  JoyNewReg = bx00001111;

  if ((IoRegisters[0xFF00] & bx00100000) != 0) { /* P14 selected (bit 4 is low) -> down/up/left/right */
      if (KeyState.Down != 0) JoyNewReg &= bx11110111; /* if DOWN -> Set P13 */
      if (KeyState.Up != 0) JoyNewReg &= bx11111011; /* if UP -> Set P12 */
      if (KeyState.Left != 0) JoyNewReg &= bx11111101; /* if LEFT -> Set P11 */
      if (KeyState.Right != 0) JoyNewReg &= bx11111110; /* if RIGHT -> Set P10 */
    } else if ((IoRegisters[0xFF00] & bx00010000) != 0) {  /* P15 selected (bit 5 is low) -> Start/Select/B/A */
      if (KeyState.Start != 0) JoyNewReg &= bx11110111; /* if ENTER -> Set P13 */
      if (KeyState.Select != 0) JoyNewReg &= bx11111011; /* if TAB -> Set P12 */
      if (KeyState.A != 0) JoyNewReg &= bx11111101; /* if CTRL -> Set P11 */
      if (KeyState.B != 0) JoyNewReg &= bx11111110; /* if ALT -> Set P10 */
    } else if ((IoRegisters[0xFF00] & bx00110000) == 0) {  /* no bits 4 & 5 set, maybe the game wants to get the system type... */
      if (IoRegisters[0xFF00] == 3) IoRegisters[0xFF00] = 0xF0; /* returns FXh to indicate a classic GameBoy device */
  }

  IoRegisters[0xFF00] = ((IoRegisters[0xFF00] & bx11110000) | JoyNewReg);   /* update the joypad register [FF00h] */

  if (JoyOldReg != JoyNewReg) JoyOldReg = JoyNewReg; /* If a key has been pressed/released update the JoyOldReg variable */

}
