/*
  zBoy - A GameBoy emulator for Windows and Linux
  Copyright (C) Mateusz Viste 2010, 2011, 2012

  ----------------------------------------------------------------------
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
  ----------------------------------------------------------------------
*/


#ifdef __zboy4linux__
  #pragma message("Compiling zBoy for Linux...")
#endif

#ifdef __zboy4windows__
  #pragma message("Compiling zBoy for Windows...")
  #define printf(...) /* Ignore any printf() calls on windows */
  /* void printf(char *s, ...) {
  } */
#endif


/* include some standard C headers */
#include <pthread.h>    /* for multi-threading capabilities */
#include <stdio.h>      /* for printf() and fprintf() */
#include <stdlib.h>     /* for atoi() and exit(), random() and randomize() */
#include <time.h>       /* for time-related functions (clock, time...) */
#include <stdint.h>     /* for uint8_t type */
#include <signal.h>
#include "SDL/SDL.h"

#define pVer "0.52"
#define pDate "2010,2011,2012"

/*  The options below make zBoy to be compiled in debug mode, and optionally trace CPU activity to a trace file */
/* #define DEBUGMODE */
/* #define DEBUGTRACEFILE */
/* #define MEMDUMP_AT_EXIT */

/* Produce a stand-alone exe file with an embedded ROM file */
/* #define EMBEDROM "embedrom/dracula.c" */
/* #define EMBEDROM "embedrom/pacman.c" */
/* #define EMBEDROM "embedrom/drmario.c" */

#ifndef DEBUGMODE   /* If not in debugmode, then disable PrintDebug */
  /* #define PrintDebug printf */
  /* #define PrintDebug(...) */
  void PrintDebug(char *s, ...) {
  }
  #else
  #define PrintDebug printf
#endif

#include "zboystructs.h"

struct RomInformations RomInfos;

char CartTypes[256][256];
signed char QuitEmulator;
signed int AskForScreenshot = 0;
unsigned int LastScreenshotTime = 0;
uint8_t NetPlay = 0;
char *NetPlayPeer;
char *NetPlayBindaddr = NULL;

unsigned int GraphicWidth, GraphicHeight;
char UserMessage[1024];
unsigned int UserMessageFramesLeft = 0;

/* the zboyversion() function is used by many modules to display the app version */
char *zboyversion() {
  return(pVer);
}
/* the zboyversion() function is used by many modules to display the app date */
char *zboydate() {
  return(pDate);
}

#include "about.h"
#include "declares.h"         /* Some SUB/FUNCTION declarations */
#include "binary.h"           /* Provides basic support to binary notation (eg. bx01100111) */
#include "adjtiming.c"        /* AdjustTiming() */
#include "config.h"           /* zBoy_loadconf() & zBoy_saveconf() */
#include "numconv.c"          /* UbyteToByte() and DwordVal() */
#include "presskey.c"         /* PressAnyKey() */
#include "getsavedir.h"       /* Provides the GetSaveDir() call */
#include "getsnapshotfilename.c"
#include "mmu.c"              /* Include Memory Management Unit emulation core */
#include "hiscores.c"         /* Provides SaveHiScoresMem() and LoadHiScoresMem() */
#include "video.c"            /* Video emulation core */
#include "loadpal.h"          /* LoadPalette() */
#include "joypad.c"           /* Joypad interface emulation */
#include "cpu-timer.c"        /* Include timer emulation */
#include "cpu-divider.c"      /* Include divider emulation */
#ifdef DEBUGMODE
  #include "debug.c"          /* Includes some debugging routines */
#endif
#include "cpu-z80.c"          /* Z80 CPU emulation core */
#include "battery.c"          /* Provides LoadBattRAM() and SaveBattRAM() */
#include "font.c"             /* Includes font data used by zBoy to display onscreen messages */
#include "wordwrap.h"         /* WordWrap(), required by PrintMsg() */
#include "printmsg.c"         /* PrintMsg(text AS STRING) */
#include "loadrom.h"          /* LoadROM(RomFile AS STRING) */
#include "serial.c"           /* Serial interface emulation */
#include "savepcxfile.c"      /* Screenshot routine SavePcxFile() */
#include "save.c"             /* SaveGame() and LoadGame() */
#include "reset.c"            /* ResetEmulator() */
#include "ssaver.c"           /* InhibitScreenSaver() */
#include "about.c"            /* printhelp() */

#ifdef ZBOY_GTK_GUI
  #include "zboygui.h"
#endif


int main (int argc, char *argv[]) {
  char *NoYes[2] = {"No", "Yes"};
  struct zboyparamstype zboyparams;
  int skipgui;
  int x;
  int UsedCycles;
  /* Prepare SDL-related variables */
  SDL_Surface *screen;      /* virtual SDL screen area */
  SDL_Joystick *joystick;   /* joystick handler */
  SDL_Color palcolors[256]; /* SDL palette */
  /* double LastDebugOnscreen, OneSecondPollingTimer, LastScreenSaverPoke; */
  pthread_t thread_InhibitScreenSaver;

  /* Prepare thread attribute for threads that we'll launch later */
  pthread_attr_t threadattr;                                         /* This is *mandatory* to run */
  pthread_attr_init(&threadattr);                                    /* threads as detached. Otherwise */
  pthread_attr_setdetachstate(&threadattr, PTHREAD_CREATE_DETACHED); /* you're going for massive mem leaks. */

  /* Set zBoy parameters to default values */
  zboyparams.ShowRomInfos = 0;
  zboyparams.NoScreenSaver = 0;
  zboyparams.ShowFPS = 0;
  zboyparams.NoSpeedLimit = 0;
  zboyparams.HiScoresMem = 0;
  zboyparams.NoCpuIdle = 0;
  zboyparams.scalingalgo = REFRESHSCREEN_2X;  /* By default, use Scale2x scaling... */
  zboyparams.GraphicScaleFactor = 4;          /* ...with 4x magnification. */
  zboyparams.fpslimit = 60;
  zboyparams.bootlogo = 0;
  zboyparams.palette = 0; /* Set up the default palette */
  zboyparams.colorization = 1;
  RomInfos.MemoryROM_PTR = MemoryROM;

  /* Here I am preloading a default configuration for input keys mapping */
  zboyparams.key_up = keystring2int("kb.up");
  zboyparams.key_down = keystring2int("kb.down");
  zboyparams.key_left = keystring2int("kb.left");
  zboyparams.key_right = keystring2int("kb.right");
  zboyparams.key_start = keystring2int("kb.ret");
  zboyparams.key_select = keystring2int("kb.tab");
  zboyparams.key_a = keystring2int("kb.lctrl");
  zboyparams.key_b = keystring2int("kb.lalt");
  zboyparams.key_turboa = keystring2int("none");
  zboyparams.key_turbob = keystring2int("none");
  zboyparams.key_bckg = keystring2int("kb.f1");
  zboyparams.key_sprt = keystring2int("kb.f2");
  zboyparams.key_wind = keystring2int("kb.f3");
  zboyparams.key_save = keystring2int("kb.f5");
  zboyparams.key_load = keystring2int("kb.f7");
  zboyparams.key_asht = keystring2int("kb.f8");
  zboyparams.key_shot = keystring2int("kb.f9");
  zboyparams.key_rset = keystring2int("kb.f10");
  zboyparams.key_quit = keystring2int("kb.esc");

  zBoy_loadconfig(&zboyparams); /* (try to) load parameters from the zBoy configuration file */

  /*printf("Parsing command-line parameters...\n"); */

  for (x = 1; x < argc; x++) {  /* Here I am parsing command-line parameters */
    /*printf("Argument #%d: %s\n", x, argv[x]); */
    if (strcasecmp(argv[x], "--SCALE") == 0) {
        x++;  /* go to next param */
        if (argc > x) {
          if ((atoi(argv[x]) < 1) || (atoi(argv[x]) > 8)) {
              zboyparams.GraphicScaleFactor = 4;    /* Fallback to x4 */
              RomInfos.Filename[0] = 0;  /* Clear out filename, to make sure to display usage screen later */
              x = 250;  /* make sure to quit parsing now */
            } else if (atoi(argv[x]) == 1) {
              zboyparams.GraphicScaleFactor = 1;    /* use 1x scaling */
              zboyparams.scalingalgo = REFRESHSCREEN_NOSCALE; /* No scaling at all */
            } else {
              zboyparams.GraphicScaleFactor = atoi(argv[x]);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_BASICSCALE; /* Nearest Neighbor scaling */
          }
        } else {
          RomInfos.Filename[0] = 0;
        }
    } else if (strcasecmp(argv[x], "--help") == 0) {
      printhelp();
      return(0);
    } else if (strcasecmp(argv[x], "--colorization") == 0) {
      zboyparams.colorization = 1;
    } else if (strcasecmp(argv[x], "--nocolorization") == 0) {
      zboyparams.colorization = 0;
    } else if (strcasecmp(argv[x], "--SCALE2X") == 0) {
        x++;  /* go to next param */
        if (argc > x) {
          if ((atoi(argv[x]) != 2) && (atoi(argv[x]) != 4) && (atoi(argv[x]) != 6) && (atoi(argv[x]) != 8)) {
              zboyparams.GraphicScaleFactor = 4;     /* Fallback to x4 */
              RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
              x = 250;  /* make sure to quit parsing now */
            } else {
              zboyparams.GraphicScaleFactor = atoi(argv[x]);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_2X;
          }
        } else {
          RomInfos.Filename[0] = 0;
        }
    } else if (strcasecmp(argv[x], "--SCALE3X") == 0) {
        x++;  /* go to next param */
        if (argc > x) {
          if ((atoi(argv[x]) != 3) && (atoi(argv[x]) != 6)) {
              zboyparams.GraphicScaleFactor = 4;     /* Fallback to x4 */
              RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
              x = 250;  /* make sure to quit parsing now */
            } else {
              zboyparams.GraphicScaleFactor = atoi(argv[x]);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_3X; /* aka AdvMAME3x */
          }
        } else {
          RomInfos.Filename[0] = 0;
        }
    } else if (strcasecmp(argv[x], "--EAGLESCALE") == 0) {
        x++;  /* go to next param */
        if (argc > x) {
          if ((atoi(argv[x]) != 2) && (atoi(argv[x]) != 4) && (atoi(argv[x]) != 6) && (atoi(argv[x]) != 8)) {
              zboyparams.GraphicScaleFactor = 4;     /* Fallback to x4 */
              RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
              x = 250;  /* make sure to quit parsing now */
            } else {
              zboyparams.GraphicScaleFactor = atoi(argv[x]);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_EAGLE;
          }
        } else {
          RomInfos.Filename[0] = 0;
        }
    } else if (strcasecmp(argv[x], "--FPSLIMIT") == 0) {
        x++;  /* go to next param */
        if (argc > x) {
          if ((atoi(argv[x]) < 1) || (atoi(argv[x]) > 60)) {
            zboyparams.fpslimit = 60;  /* Fallback to 60 (no frameskip) */
            RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
            x = 250;  /* make sure to quit parsing now */
          } else {
            zboyparams.fpslimit = atoi(argv[x]);  /* retrieve provided skipframes parameter */
          }
        } else {
          RomInfos.Filename[0] = 0;
        }
    } else if (strcasecmp(argv[x], "--BOOTLOGO") == 0) {   /* Enable Gameboy's bootstrap */
        zboyparams.bootlogo = 1;
    } else if (strcasecmp(argv[x], "--PALETTE") == 0) {   /* load a custom palette */
        x++;  /* go to next param */
        if (argc > x) {
          if (strcasecmp(argv[x], "gbmicro") == 0) zboyparams.palette = 0;
          if (strcasecmp(argv[x], "peasoup") == 0) zboyparams.palette = 1;
        } else {
          RomInfos.Filename[0] = 0;
        }
    } else if (strcasecmp(argv[x], "--ROMINFOS") == 0) {  /* shows various informations about the ROM before running it. */
        zboyparams.ShowRomInfos = 1;
    } else if (strcasecmp(argv[x], "--NOCPUIDLE") == 0) {  /* do not attempt to idle CPU when the emulator have some spare time. */
        zboyparams.NoCpuIdle = 1;
    } else if (strcasecmp(argv[x], "--INPUT") == 0) {
      if (argc < (x + 10)) {
           RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
           x = 250;  /* make sure to quit parsing now */
        } else {
          zboyparams.key_up = keystring2int(argv[x + 1]);
          zboyparams.key_down = keystring2int(argv[x + 2]);
          zboyparams.key_left = keystring2int(argv[x + 3]);
          zboyparams.key_right = keystring2int(argv[x + 4]);
          zboyparams.key_start = keystring2int(argv[x + 5]);
          zboyparams.key_select = keystring2int(argv[x + 6]);
          zboyparams.key_a = keystring2int(argv[x + 7]);
          zboyparams.key_b = keystring2int(argv[x + 8]);
          zboyparams.key_turboa = keystring2int(argv[x + 9]);
          zboyparams.key_turbob = keystring2int(argv[x + 10]);
          x += 10;
      }
    } else if (strcasecmp(argv[x], "--FNKEYS") == 0) {
      if (argc < (x + 9)) {
           RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
           x = 250;  /* make sure to quit parsing now */
        } else {
          zboyparams.key_bckg = keystring2int(argv[x + 1]);
          zboyparams.key_sprt = keystring2int(argv[x + 2]);
          zboyparams.key_wind = keystring2int(argv[x + 3]);
          zboyparams.key_save = keystring2int(argv[x + 4]);
          zboyparams.key_load = keystring2int(argv[x + 5]);
          zboyparams.key_asht = keystring2int(argv[x + 6]);
          zboyparams.key_shot = keystring2int(argv[x + 7]);
          zboyparams.key_rset = keystring2int(argv[x + 8]);
          zboyparams.key_quit = keystring2int(argv[x + 9]);
          x += 9;  /* Redefinition of keys successful, let's continue parsing... */
      }
    } else if (strcasecmp(argv[x], "--NOSPEEDLIMIT") == 0) {
        zboyparams.NoSpeedLimit = 1;
    } else if (strcasecmp(argv[x], "--NOSCREENSAVER") == 0) {
        zboyparams.NoScreenSaver = 1;
    } else if (strcasecmp(argv[x], "--SHOWFPS") == 0) {
        zboyparams.ShowFPS = 1;
    } else if (strcasecmp(argv[x], "--HISCORESMEM") == 0) {
        zboyparams.HiScoresMem = 1;
    } else if (strcasecmp(argv[x], "--NETPLAY") == 0) {
        x += 1;
        if (x < argc) {
            NetPlay = 1; /* atoi(argv[x]); */
            NetPlayPeer = argv[x];
          } else {
            RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
        }
    } else if (strcasecmp(argv[x], "--BINDADDR") == 0) {
        x += 1;
        if (x < argc) {
            NetPlayBindaddr = argv[x];
          } else {
            RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
        }
    } else {  /* anything else is assumed to be a filename */
        if (argv[x][0] != 0) { /* ignore if the parameter is empty */
          if (RomInfos.Filename[0] == 0) {
              snprintf(RomInfos.Filename, 1022, "%s", argv[x]);
            } else {
              RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
              x = 250;  /* make sure to quit parsing now */
          }
        }
    }
  }

  if (RomInfos.Filename[0] != 0) skipgui = 1; else skipgui = 0;
  #ifndef ZBOY_GTK_GUI
  skipgui = 1;
  #endif
  do { /* start looping to get back to the GUI (only if GUI enabled at compile time) */
  /* if GUI is compiled in, and no filename have been provided on command line, then run the GUI */
  if (skipgui == 0) {
    #ifdef ZBOY_GTK_GUI
    int guiresult = zboygui(&RomInfos, &zboyparams);
    if (guiresult == 1) return(0); /* if GUI has been closed, then shut down everything */
    #endif
  }
  zBoy_fixconfig(&zboyparams);  /* check the configuration, and fix it if needed */
  switch (zboyparams.palette) {
    case 1:
      LoadPalette("peasoup", ScreenPalette32); /* Load the palette */
      break;
    default:
      LoadPalette("gbmicro", ScreenPalette32); /* Load the palette */
      break;
  }
  /* Set up input keys definitions */
  if (RedefineInputKeys(int2keystring(zboyparams.key_up), int2keystring(zboyparams.key_down), int2keystring(zboyparams.key_left), int2keystring(zboyparams.key_right), int2keystring(zboyparams.key_start), int2keystring(zboyparams.key_select), int2keystring(zboyparams.key_a), int2keystring(zboyparams.key_b), int2keystring(zboyparams.key_turboa), int2keystring(zboyparams.key_turbob)) != 0) printf("error redefining input keys\n");;
  if (RedefineFunctionKeys(int2keystring(zboyparams.key_bckg), int2keystring(zboyparams.key_sprt), int2keystring(zboyparams.key_wind), int2keystring(zboyparams.key_save), int2keystring(zboyparams.key_load), int2keystring(zboyparams.key_asht), int2keystring(zboyparams.key_shot), int2keystring(zboyparams.key_rset), int2keystring(zboyparams.key_quit)) != 0) printf("error redefining function keys\n");

  /* init the screen via the whole SDL mess */
/*  #ifdef DEBUGMODE
    GraphicWidth = 800;
    GraphicHeight = 900;
    GraphicScaleFactor = 2;
  #else */
    GraphicWidth = 160 * zboyparams.GraphicScaleFactor;
    GraphicHeight = 144 * zboyparams.GraphicScaleFactor;
/*  #endif */

  /* Initialize the SDL library... */
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0) {
    printf("Couldn't initialize SDL: %s\n", SDL_GetError());
    return(1);
  }
  SDL_ShowCursor(SDL_DISABLE); /* Disable (hide) mouse cursor */
  atexit(SDL_Quit); /* ...and remember to deallocate SDL before quitting. */
  joystick = SDL_JoystickOpen(0);  /* ...and open the 1st system joystick */

  /* if (FullscreenFlag != 0) {
      switch (GraphicScaleFactor) {
        case 1:
          GraphicWidth = 320;
          GraphicHeight = 200;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 27;    // add 27 pixels of border on the top of the screen to vertically center the display area
          break;
        case 2:
          GraphicWidth = 512;
          GraphicHeight = 384;
          ScreenBorderRight = 96;  // add 96 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 47;    // add 47 pixels of border on the top of the screen to vertically center the display area
          break;
        case 3:
          GraphicWidth = 640;
          GraphicHeight = 480;
          GraphicScaleFactor = 3;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 23;    // add 24 pixels of border on the top of the screen to vertically center the display area
          break;
        case 4:
          GraphicWidth = 800;
          GraphicHeight = 600;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 11;    // add 11 pixels of border on the top of the screen to vertically center the display area
          break;
        case 5:
          GraphicWidth = 1024;
          GraphicHeight = 768;
          ScreenBorderRight = 112; // add 112 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 23;    // add 23 pixels of border on the top of the screen to vertically center the display area
          break;
        case 6:
          GraphicWidth = 1280;
          GraphicHeight = 1024;
          ScreenBorderRight = 160; // add 160 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 80;    // add 80 pixels of border on the top of the screen to vertically center the display area
          break;
        case 7:
          GraphicWidth = 1280;
          GraphicHeight = 1024;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 8;     // add 23 pixels of border on the top of the screen to vertically center the display area
          break;
      }

      // Set up the fullscreen video mode (some available flags: SDL_SWSURFACE, SDL_DOUBLEBUF, SDL_HWSURFACE, SDL_FULLSCREEN...)
      screen = SDL_SetVideoMode(GraphicWidth, GraphicHeight, 8, SDL_SWSURFACE);
      if (screen == NULL) {
        printf("Couldn't set video mode: %s\n", SDL_GetError());
        return(1);
      }
    } else { */
      /* Set up a video mode (that is, create a window on screen) */
      screen = SDL_SetVideoMode(GraphicWidth, GraphicHeight, 8, SDL_SWSURFACE);
      if (screen == NULL) {
        printf("Couldn't set video mode: %s\n", SDL_GetError());
        return(1);
      }
  /* } */
  SDL_WM_SetCaption("zBoy", "zBoy");
  /* On Windows, use a dirty trick to hide that ugly title-bar icon... */
  #ifdef __zboy4windows__
  do {
    SDL_Surface *dummyicon;
    dummyicon = SDL_CreateRGBSurface(SDL_SWSURFACE, 32, 32, 32, 0xFF000000, 0x00FF0000, 0x0000FF00, 0xFFFFFFFF);
    SDL_WM_SetIcon(dummyicon, NULL);
  } while(0);
  #endif

  InitScreenOldBuffer();   /* Init the screen buff used for LCD change detection (will force the complete draw of the 1st frame) */

  /* Put palette into SDL now, in case we need to display an error message */
  for (x = 0; x < 256; x++) {
    palcolors[x].r = (ScreenPalette32[x] >> 16) & 0xFF;
    palcolors[x].g = (ScreenPalette32[x] >> 8) & 0xFF;
    palcolors[x].b = ScreenPalette32[x] & 0xFF;
  }
  SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, palcolors, 0, 255);

  /* if the GUI is not active, then load the ROM file */
  if (skipgui != 0) {
    #ifndef EMBEDROM
    if (RomInfos.Filename[0] == 0) {
      printf("No ROM filename provided!\n");
      sprintf(UserMessage, "\n  *** ZBOY V%s ***\n\nUSAGE: ZBOY FILE.GB\n\nSEE ZBOY.TXT FOR DETAILED INSTRUCTIONS.", pVer);
      PrintMsg(UserMessage, 0);
      RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
      PressAnyKey();
      return(0);
    }
    printf("Loading the ROM data into memory...\n");
    x = LoadRom(&RomInfos);
    if (x != 0) {
      char *errstr = LoadRomErrStr(x);
      printf("Error: Could not load the ROM file. [%d]\n", x);
      PrintMsg(errstr, 0);
      RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
      PressAnyKey();
      return(0);
    }
    #endif
  }

  if (zboyparams.colorization != 0) CheckForColorizedPalette(RomInfos.CrcSum , ScreenPalette32);
  printf("Uploading palette into the virtual screen...\n");
  for (x = 0; x < 256; x++) {
    palcolors[x].r = (ScreenPalette32[x] >> 16) & 0xFF;
    palcolors[x].g = (ScreenPalette32[x] >> 8) & 0xFF;
    palcolors[x].b = ScreenPalette32[x] & 0xFF;
  }
  SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, palcolors, 0, 255);

  if (NetPlay != 0) {
    if ((s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) printf("socket problem\n");
    memset((char *) &si_me, 0, sizeof(si_me));
    si_me.sin_family = AF_INET;
    /* if (NetPlay == 1) si_me.sin_port = htons(8765); else si_me.sin_port = htons(8764); */
    si_me.sin_port = htons(8764);
    if (NetPlayBindaddr == NULL) {
        si_me.sin_addr.s_addr = htonl(INADDR_ANY); /* listen on all addresses if no bind specified */
      } else {
        if (inet_aton(NetPlayBindaddr, &si_me.sin_addr) == 0) printf("inet_aton() failed - bindaddr is probably not a valid IP\n");
    }
    memset((char *) &si_other, 0, sizeof(si_other));
    si_other.sin_family = AF_INET;
    /* if (NetPlay == 1) si_other.sin_port = htons(8764); else si_other.sin_port = htons(8765); */
    si_other.sin_port = htons(8764);
    /* if (inet_aton("127.0.0.1", &si_other.sin_addr)==0) printf("inet_aton() failed - NetPlay peer is probably not a valid IP\n"); */
    if (inet_aton(NetPlayPeer, &si_other.sin_addr) == 0) printf("inet_aton() failed - NetPlay peer is probably not a valid IP\n");
    if (bind(s, (struct sockaddr *) &si_me, sizeof(si_me)) == -1) printf("bind() problem\n");
  }

  /* printf("Loading ROM informations...\n"); */
  /* LoadRomInfos(&RomInfos); */

  if (zboyparams.ShowRomInfos != 0) {
    char messagetoprint[1024];
    sprintf(messagetoprint, "TITLE: %s\n\nCRC32: %08X\n\nLICENSEE: %s\n\nCART TYPE: %s\n\nBATTERY: %s\n\nROM SIZE: %u KIB\n\nRAM SIZE: %u KIB\n\nPRESS ANY KEY...", RomInfos.NiceTitle, RomInfos.CrcSum, RomInfos.Licensee, RomInfos.CartTypeString, NoYes[RomInfos.Battery], RomInfos.RomSize >> 10, RomInfos.RamSize >> 10);
    /* Switch the string to all-uppercase */
    for (x = 0; messagetoprint[x] != 0; x++) if ((messagetoprint[x] >= 'a') && (messagetoprint[x] <= 'z')) messagetoprint[x] -= 32;
    PrintMsg(messagetoprint, 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    printf("Title......: %s (\"%s\")\n", RomInfos.Title, RomInfos.NiceTitle);
    printf("Filename...: %s\n", RomInfos.Filename);
    printf("CRC32 sum..: %8X\n", RomInfos.CrcSum);
    printf("Licensee...: %s\n", RomInfos.Licensee);
    if (RomInfos.ColorGB == 2) {
        printf("ColorGB....: Yes (GBC ONLY)\n");
      } else {
        printf("ColorGB....: %s\n", NoYes[RomInfos.ColorGB]);
    }
    printf("SuperGB....: %s\n", NoYes[RomInfos.SuperGB]);
    printf("Cart Type..: %s\n", RomInfos.CartTypeString);
    printf("Battery....: %s\n", NoYes[RomInfos.Battery]);
    if (RomInfos.Destination == 0) {
        printf("Country....: Japan\n");
      } else {
        printf("Country....: Other (not Japan)\n");
    }
    printf("ROM Size...: %u KiB\n", (RomInfos.RomSize >> 10));
    printf("RAM Size...: %u KiB\n", (RomInfos.RamSize >> 10));
    PressAnyKey();
  }

  if (RomInfos.Rumble != 0) {
    PrintMsg("THIS ROM USES A RUMBLE MECHANISM. ZBOY DOESN'T SUPPORT THAT YET! SORRY.", 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    PressAnyKey();
    return(0);
  }
  if (RomInfos.TimerRTC != 0) {
    PrintMsg("THIS ROM USES AN INTERNAL RTC TIMER CLOCK. ZBOY DOESN'T SUPPORT SUCH ROMS YET! SORRY.", 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    PressAnyKey();
    return(0);
  }
  /*
  IF RomInfos.SuperGB = 1 THEN PRINT : PRINT "Error: SuperGB ROMs are not supported (yet). Sorry." : SLEEP : END
  IF RomInfos.ColorGB = 2 THEN PRINT : PRINT "Error: This ROM requires a GameBoy Color emulator. zBoy supports only legacy GameBoy. Sorry." : SLEEP : END
  */

  switch (RomInfos.MbcModel) {  /* Check the MBC chip type, and select the right Read/Write functions */
    case 0:   /* ROM ONLY */
      MemoryWriteSpecial = &MBC0_MemoryWrite;
      MemoryReadSpecial = &MBC0_MemoryRead;
      break;
    case 1:   /* MBC1 */
      MemoryWriteSpecial = &MBC1_MemoryWrite;
      MemoryReadSpecial = &MBC1_MemoryRead;
      break;
    case 2:   /* MBC2 */
      MemoryWriteSpecial = &MBC2_MemoryWrite;
      MemoryReadSpecial = &MBC2_MemoryRead;
      break;
    case 3:   /* MBC3 */
      MemoryWriteSpecial = &MBC3_MemoryWrite;
      MemoryReadSpecial = &MBC3_MemoryRead;
      break;
    case 5:   /* MBC5 */
      MemoryWriteSpecial = &MBC5_MemoryWrite;
      MemoryReadSpecial = &MBC5_MemoryRead;
      break;
    default:  /* Any other (not supported) memory controller */
      PrintMsg("ERROR: THE MEMORY CONTROLLER USED BY THIS ROM IS NOT SUPPORTED (YET) BY ZBOY. SORRY.", 0);
      RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
      PressAnyKey();
      return(0);
      break;
  }

  /* Set up the window's title */
  if (RomInfos.NiceTitle[0] != 0) {
      snprintf(zboyparams.windowtitle, 256, "%s (zBoy)", RomInfos.NiceTitle);
    } else {
      snprintf(zboyparams.windowtitle, 256, "zBoy");
  }
  SDL_WM_SetCaption(zboyparams.windowtitle, "zBoy");

  srand(time(NULL));  /* Seed the random generator */

  if (zboyparams.NoScreenSaver != 0) {  /* If ScreenSaver inhibition is turned on, launch the InhibitScreenSaver thread */
    if (pthread_create(&thread_InhibitScreenSaver, &threadattr, &InhibitScreenSaver, NULL) != 0) printf("Error: Inhibit ScreenSaver thread_create failed.\n");
  }

  /*InitRAM() */    /* Init RAM areas to random values (as in a real GameBoy) */
  if ((RomInfos.Battery != 0) && (RomInfos.RamSize > 0)) LoadBattRAM();  /* check for battery-saved RAM */

  BootRomEnabledFlag = zboyparams.bootlogo; /* Set up the boot logo flag */
  ResetEmulator(); /* Fire up the Zilog80 monster, and init all required variables */

  if (zboyparams.HiScoresMem != 0) LoadHiScoresMem();   /* Check out if need to look after hiscores */

  printf("Starting emulation...\n");

  do {
    UsedCycles = CpuExec();
    TotalCycles += UsedCycles;    /* Increment the global cycles counter */
    uTimer(UsedCycles);           /* Update uTimer */
    incDivider(UsedCycles);       /* Increment the DIV register */
    VideoSysUpdate(UsedCycles, screen, &zboyparams);   /* Update Video subsystem */

    CheckSerialLink(UsedCycles);  /* Check serial link (not supported yet) */
    CheckJoypad(UsedCycles, &zboyparams);  /* Update the Joypad register */
    CheckInterrupts();  /* Check if any interrupt has been requested since last time */
/*    #ifdef DEBUGMODE
      if ((TIMER - LastDebugOnscreen) > 0.5) { */
/*        DebugSpriteDump(1); */  /* perform some debug dumps onscreen (OBJ table at &h8000) */
/*        DebugSpriteDump(2); */  /* perform some debug dumps onscreen (OBJ table at &h8800) */
/*        DebugPaletteDump(); */  /* Dump system's palette onscreen */
/*        DebugShowCPUspeed(); */ /* Calc current CPU speed and put it onscreen */
/*        DebugShowCpuRegisters(); */ /* Display state of CPU registers */
/*        DebugOnscreen(); */     /* Display debug buffer informations onscreen */
/*        LastDebugOnscreen = TIMER;
      }
    #endif */
  } while (QuitEmulator == 0);

  if ((RomInfos.Battery != 0) && (RomInfos.RamSize > 0)) SaveBattRAM();  /* check for battery-saved RAM */
  if (zboyparams.HiScoresMem != 0) SaveHiScoresMem();

  SDL_JoystickClose(joystick);

  #ifdef MEMDUMP_AT_EXIT
    DIM AS UINTEGER TempUINT             /* ----------------- */
    OPEN "memdump.bin" FOR OUTPUT AS #1  /* This is a memory  */
    FOR TempUINT = &h0000 TO &hFFFF      /* dump routine.     */
      MemoryRead(TempUINT, x)            /*     --  --        */
      PUT #1,, x                         /* Don't include in  */
    NEXT TempUINT                        /* official release! */
    CLOSE #1                             /* ----------------- */
  #endif

  #ifdef DEBUGMODE
    if (QuitEmulator != 0) {
      DebugOnScreen();
      PressAnyKey();
    }
  #endif

  if (NetPlay != 0) close(s);

  #ifdef NETPLAYSUPPORT
    IF (NetPlay = 1) THEN  /* Close net stuff if NetPlay is used */
      NeedToCloseNetworkThreads = 1
      ThreadWait(NetPollingThread)
      ThreadWait(NetSendingThread)
    END IF
  #endif

  SDL_Quit();
  fflush(stdin); /* Flush keyb buffer */
  QuitEmulator = 0;

  } while(skipgui == 0);  /* reloop into the GUI if GUI is compiled in, and if zBoy is not launched headless... */

  return(0);
}
